# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Blenderbot Checkpoint Converter."""

from __future__ import annotations

import math
from typing import Any, Dict, List, cast

import numpy as np
import torch
from transformers import BlenderbotConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    ENCODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    EncoderDecoderConverter,
)
from friendli.modules.converter.schema import ConvertInfo


class BlenderbotConverter(EncoderDecoderConverter):
    """BlenderbotForConditionalGeneration Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Blenderbot architectures's config can be converted to Friendli format."""
        super().check_config()
        config = cast(BlenderbotConfig, self.config)
        try:
            if config.activation_function not in SUPPORTED_GELU_FAMILY:
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(BlenderbotConfig, self.config).activation_function}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if not config.tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=False'",
                    valid_options=[True],
                )
            if self.encoder_num_attention_heads != self.decoder_num_attention_heads:
                raise NotSupportedCheckpointError(
                    invalid_option=(
                        f"encoder_num_attention_heads={self.encoder_num_attention_heads} "
                        f"decoder_num_attention_heads={self.decoder_num_attention_heads}"
                    ),
                    valid_options=[
                        "encoder_num_attention_heads == decoder_num_attention_heads"
                    ],
                )
            if config.decoder_ffn_dim != config.encoder_ffn_dim:
                raise NotSupportedCheckpointError(
                    invalid_option=(
                        f"encoder_ffn_dim={config.encoder_ffn_dim} "
                        f"decoder_ffn_dim={config.decoder_ffn_dim}"
                    ),
                    valid_options=["encoder_ffn_dim == decoder_ffn_dim"],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def token_embed_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """Reshape token embedding weight for Blenderbot's embedding layer."""
        assert len(params) == 1
        embed_dim = cast(BlenderbotConfig, self.config).d_model
        embed_scale = (
            math.sqrt(embed_dim)
            if cast(BlenderbotConfig, self.config).scale_embedding
            else 1.0
        )
        embed_weight = params[0]
        embed_weight = embed_weight * embed_scale
        return embed_weight

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(BlenderbotConfig, self.config)

        logger.warn(
            "Since Blenderbot uses absolute position embedding, 'max_input_length' and "
            "'max_output_length' cannot be larger than %d.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        decoder_start_token_id = self.get_decoder_start_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.encoder_head_size,
            "num_heads": self.encoder_num_attention_heads,
            "hidden_size": self.encoder_hidden_size,
            "ff_intermediate_size": self.decoder_ff_intermediate_size,
            "num_encoder_layers": self.encoder_layer_num,
            "num_decoder_layers": self.decoder_layer_num,
            "max_input_length": config.max_position_embeddings,
            "max_output_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "decoder_start_token": (
                decoder_start_token_id
                if decoder_start_token_id is not None
                else "FILL ME"
            ),
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "blenderbot"

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in Blenderbot."""
        return [
            ConvertInfo(
                param_names=["model.shared.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.shared.weight"],
                data_type=self.data_type,
                converted_name="head_fc/weight:0",
                reshape_fn=self.head_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.encoder.embed_positions.weight"],
                data_type=self.data_type,
                converted_name=f"{ENCODER_PREFIX}/wpe/weight:0",
                reshape_fn=self.pos_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.embed_positions.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/wpe/weight:0",
                reshape_fn=self.pos_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.encoder.layer_norm.weight"],
                data_type=self.data_type,
                converted_name=f"{ENCODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.encoder.layer_norm.bias"],
                data_type=self.data_type,
                converted_name=f"{ENCODER_PREFIX}/ln_f/beta:0",
                reshape_fn=self.ln_bias_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.layer_norm.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.layer_norm.bias"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/beta:0",
                reshape_fn=self.ln_bias_reshape,
            ),
        ]

    @property
    def encoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in Blenderbot's encoder."""
        convert_info_list = []
        for i in range(self.encoder_layer_num):
            layer_prefix = f"{self.encoder_layer_prefix}{i}."
            converted_prefix = f"{ENCODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.weight",
                            f"{layer_prefix}self_attn.k_proj.weight",
                            f"{layer_prefix}self_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.bias",
                            f"{layer_prefix}self_attn.k_proj.bias",
                            f"{layer_prefix}self_attn.v_proj.bias",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/bias:0",
                        reshape_fn=self.qkv_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in Blenderbot's decoder."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.weight",
                            f"{layer_prefix}self_attn.k_proj.weight",
                            f"{layer_prefix}self_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.bias",
                            f"{layer_prefix}self_attn.k_proj.bias",
                            f"{layer_prefix}self_attn.v_proj.bias",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/bias:0",
                        reshape_fn=self.qkv_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}encoder_attn_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}encoder_attn_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_3/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_3/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}encoder_attn.q_proj.weight",
                            f"{layer_prefix}encoder_attn.k_proj.weight",
                            f"{layer_prefix}encoder_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}cross_attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}encoder_attn.q_proj.bias",
                            f"{layer_prefix}encoder_attn.k_proj.bias",
                            f"{layer_prefix}encoder_attn.v_proj.bias",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}cross_attn/c_attn/bias:0",
                        reshape_fn=self.qkv_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}encoder_attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}cross_attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}encoder_attn.out_proj.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}cross_attn/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def encoder_layer_prefix(self) -> str:
        """The layer name prefix used before Blenderbot encoder's transformer block number."""
        return "model.encoder.layers."

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before Blenderbot decoder's transformer block number."""
        return "model.decoder.layers."

    @property
    def encoder_layer_num(self) -> int:
        """The number of transformer blocks in Blenderbot encoder."""
        return cast(BlenderbotConfig, self.config).encoder_layers

    @property
    def encoder_hidden_size(self) -> int:
        """The hidden size of Blenderbot encoder."""
        return cast(BlenderbotConfig, self.config).d_model

    @property
    def encoder_num_attention_heads(self) -> int:
        """The number of attention heads of Blenderbot encoder."""
        return cast(BlenderbotConfig, self.config).encoder_attention_heads

    @property
    def encoder_head_size(self) -> int:
        """The size of each attention head of Blenderbot encoder."""
        return self.encoder_hidden_size // self.encoder_num_attention_heads

    @property
    def encoder_ff_intermediate_size(self) -> int:
        """The intermediate of the linear layer in Blenderbot encoder's MLP."""
        return cast(BlenderbotConfig, self.config).encoder_ffn_dim

    @property
    def decoder_layer_num(self) -> int:
        """The number of transformer blocks in Blenderbot decoder."""
        return cast(BlenderbotConfig, self.config).decoder_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size of Blenderbot decoder."""
        return cast(BlenderbotConfig, self.config).d_model

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads of Blenderbot decoder."""
        return cast(BlenderbotConfig, self.config).decoder_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads of blenderbot decoder."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The size of each attention head of Blenderbot decoder."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate of the linear layer in Blenderbot decoder's MLP."""
        return cast(BlenderbotConfig, self.config).decoder_ffn_dim



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli V1 Chat Completion Serving API Schemas."""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel
from typing_extensions import Required, TypedDict


class MessageParam(TypedDict, total=False):
    """Message param schema."""

    role: Required[str]
    content: Required[str]


class Message(BaseModel):
    """Message schema."""

    role: Optional[str] = None
    content: Optional[str] = None


class ChatCompletionChoice(BaseModel):
    """Chat completion choice schema."""

    index: int
    message: Message
    finish_reason: str


class ChatCompletionDeltaChoice(BaseModel):
    """Schema of chat completion choice with delta."""

    index: int
    delta: Message
    finish_reason: Optional[str] = None


class ChatCompletionUsage(BaseModel):
    """Chat completion usage schema."""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatCompletion(BaseModel):
    """Chat completion schema."""

    choices: List[ChatCompletionChoice]
    usage: ChatCompletionUsage
    created: int


class ChatCompletionLine(BaseModel):
    """Chat completion line schema."""

    choices: List[ChatCompletionDeltaChoice]
    created: int



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli SmoothQuant Quantizer Base."""

from __future__ import annotations

from abc import abstractmethod
from dataclasses import fields
from typing import Any, Dict, Iterator, List, Tuple, cast

import datasets  # type: ignore[import]
import torch

from friendli.enums import ModelDataType
from friendli.errors import NotSupportedQuantConfigError
from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.interface import ModelConversionInterface
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import get_tokenizer
from friendli.modules.quantizer.base import AbstractQuantHook, CommonQuantizer
from friendli.modules.quantizer.layers import WeightActQuantizedLinearLayer
from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import (
    ModuleName,
    QuantInput,
    TFQuantInputs,
    TFQuantResults,
    WeightActQuantResult,
)
from friendli.modules.quantizer.utils import (
    collect_stats,
    get_weight_act_quant_scales,
    quantized_linear_weight_reshape,
    quantized_qkv_weight_reshape,
    safe_load_datasets,
    scale_reshape,
)


class PreSmoother(torch.nn.Module):
    """Module for containing smoothing scale.

    This module is used to contain the smoothing scale for the quantization.
    If the matmul layer have previous layer, the smoothing scale can be migrated
    to the previous layer. But, if the matmul layer is the first layer, the scale
    need to be stored in this module. Especially, When MLP ff2 layer with previous activation
    layer that prevent migrating the scale to the previous layer needs SmoothQuant, then,
    this module is used to store the smoothing scale. [SmoothQunat Issue #15]
    (https://github.com/mit-han-lab/smoothquant/issues/15#issuecomment-1353390283).

    Args:
        in_dim (float): input dimension of the matmul layer's weight dimension.
    """

    def __init__(self, in_dim: int):
        """Initialize PreSmoother."""
        super().__init__()
        self.scale = torch.nn.Parameter(torch.ones(in_dim, dtype=torch.float32))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function of PreSmoother."""
        return (x * self.scale).to(x.dtype)


class SmoothQuantHook(AbstractQuantHook):
    """Quantization Hook for SmoothQuant."""

    @abstractmethod
    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the attention fc layer in the decoder block."""

    @abstractmethod
    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the second feed-forward layer in the decoder block."""

    @abstractmethod
    def iter_smooth_norm_weights(
        self, model: torch.nn.Module
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm and linear layer's weight per transformer block."""

    def _register_pre_smoother(self, linear: torch.nn.Linear) -> PreSmoother:
        """Register pre_smoother storing smoothing scale of linear layer."""
        pre_smoother = PreSmoother(linear.in_features).to(device=linear.weight.device)

        def pre_smoother_hook(_, x: Tuple[Any, ...]) -> Tuple[torch.Tensor, ...]:
            return (pre_smoother.forward(x[0]),)

        linear.register_forward_pre_hook(pre_smoother_hook)
        return pre_smoother

    def pre_smooth(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Pre-procedure for SmoothQuant before Smoothing."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for decoder_layer in self.get_tf_blocks(model):
            if quant_args.attn_fc_smoothing:
                attn_fc_pre_smoother = self._register_pre_smoother(
                    self.get_attn_fc_layer(decoder_layer)
                )
                decoder_layer.add_module("attn_fc_pre_smoother", attn_fc_pre_smoother)
            if quant_args.ff2_smoothing:
                ff2_pre_smoother = self._register_pre_smoother(
                    self.get_ff2_layer(decoder_layer)
                )
                decoder_layer.add_module("ff2_pre_smoother", ff2_pre_smoother)
        return model

    def sort_qkv_output_stats(self, max_output_stat: torch.Tensor) -> torch.Tensor:
        """Sort max_output_stas for seperating qkv_layer's output_stats."""
        return max_output_stat

    def copy_norms(self, model: torch.nn.Module) -> torch.nn.Module:
        """Copy and Register norms in transformer block for seperated scaling.

        In some models(e.g. llama, gptj, codegen), matmul layers share activations
        from the same norms. Therefore, we need to copy and register the norms for
        seperated smoothing scale. For example, in llama, normalization layer is
        shared with gate linear layer and attention linear layer. Thus, we need to
        copy and register the norms for each linear layer and use them for smoothing.
        """
        return model

    def get_quant_result(
        self,
        quant_inputs: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Returns the quantization result of the quantized layer.

        If the model has another quantized layer, it should be implemented in the subclass.

        """
        max_input_stats: Dict[ModuleName, torch.Tensor] = kwargs["max_input_stats"]
        max_output_stats: Dict[ModuleName, torch.Tensor] = kwargs["max_output_stats"]

        def get_scale(
            quant_input: QuantInput,
        ) -> WeightActQuantResult:
            weight, name, start, end, sort_fn = (
                quant_input.weight,
                quant_input.name,
                quant_input.start_offset,
                quant_input.end_offset,
                quant_input.sort_fn,
            )

            return get_weight_act_quant_scales(
                name,
                max_input_stats[name],
                weight[start:end],
                sort_fn(max_output_stats[name])[start:end]
                if sort_fn
                else max_output_stats[name][start:end],
            )

        return TFQuantResults(
            layer_prefix_with_index=f"{self.quantized_layer_prefix}{quant_inputs.layer_index}.",
            block=quant_inputs.block,
            q=get_scale(quant_inputs.q),
            k=get_scale(quant_inputs.k),
            v=get_scale(quant_inputs.v),
            attn_fc=get_scale(quant_inputs.attn_fc),
            ff1=get_scale(quant_inputs.ff1),
            ff2=get_scale(quant_inputs.ff2),
        )

    @property
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified modules.

        This convert_info_list is used for modules that are modified for quantization.
        Especially, for attention fc layer and MLP ff2 layer, we need to migrate
        smooth scale to the previous layer. Thus, we add the smoothing scaler, and
        modify the convert_info_list for the modified modules.

        In some models, matmul layers share activations from the same norms. Therefore,
        we use `copy_norms()` to copy and register the norms for seperated smoothing scale.
        Thus, we modify the convert_info_list for the modified modules.
        """
        sq_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        new_layer_convert_info_list = []
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"

            if sq_args.attn_fc_smoothing:
                new_layer_convert_info_list.append(
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc_pre_smoother.scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_proj/smoothquant/smoothing_vector:0",  # pylint: disable=line-too-long
                        reshape_fn=scale_reshape,
                    )
                )
            if sq_args.ff2_smoothing:
                new_layer_convert_info_list.append(
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2_pre_smoother.scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_proj/smoothquant/smoothing_vector:0",  # pylint: disable=line-too-long
                        reshape_fn=scale_reshape,
                    )
                )

        return new_layer_convert_info_list

    @property
    def quantized_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for quantized layers."""
        convert_info_list = []
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}q.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/q_weight_scale:0",  # pylint: disable=line-too-long
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}k.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/k_weight_scale:0",  # pylint: disable=line-too-long
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}v.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/v_weight_scale:0",  # pylint: disable=line-too-long
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}q.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/q_out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}k.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/k_out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}v.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/v_out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}q.in_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/in_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_proj/smoothquant/weight_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_proj/smoothquant/out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.in_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_proj/smoothquant/in_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_fc/smoothquant/weight_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_fc/smoothquant/out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.in_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_fc/smoothquant/in_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.weight_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_proj/smoothquant/weight_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.out_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_proj/smoothquant/out_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.in_scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_proj/smoothquant/in_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}q.weight",
                            f"{layer_prefix}k.weight",
                            f"{layer_prefix}v.weight",
                        ],
                        data_type=ModelDataType.INT8,
                        converted_name=f"{converted_prefix}attn/c_attn/smoothquant/weight:0",
                        reshape_fn=quantized_qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.weight"],
                        data_type=ModelDataType.INT8,
                        converted_name=f"{converted_prefix}attn/c_proj/smoothquant/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.weight"],
                        data_type=ModelDataType.INT8,
                        converted_name=f"{converted_prefix}mlp/c_fc/smoothquant/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.weight"],
                        data_type=ModelDataType.INT8,
                        converted_name=f"{converted_prefix}mlp/c_proj/smoothquant/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                ]
            )
        return convert_info_list


class SmoothQuantQuantizer(CommonQuantizer, ModelConversionInterface):
    """Quantizer for SmoothQuant."""

    def check_config(self) -> None:
        """Check if the SmoothQuant quantization config is valid."""
        quant_config = cast(SmoothQuantConfig, self.quant_config)
        smoothquant_args = quant_config.smoothquant_args
        super().check_config()
        if 0 > smoothquant_args.migration_strength > 1:
            raise NotSupportedQuantConfigError(
                invalid_option=str(smoothquant_args.migration_strength),
                valid_options=["between 0 and 1."],
            )

    def get_calib_dataset(self) -> datasets.Dataset:
        """Get calibration dataset for SmoothQuant."""
        data_cfg = self.quant_config.calibration_dataset
        tokenizer = get_tokenizer(self.converter.config.name_or_path)
        dataset = safe_load_datasets(data_cfg)

        def preprocess(example) -> Dict[str, torch.Tensor]:
            truncate_length = data_cfg.max_length * 4
            while True:
                input_ids = tokenizer(
                    example[data_cfg.lookup_column_name][:truncate_length],
                    return_tensors="pt",
                    max_length=data_cfg.max_length * 2,
                    truncation=True,
                    padding=False,
                ).input_ids

                if input_ids.size(
                    1
                ) >= data_cfg.max_length * 2 or truncate_length >= len(
                    example[data_cfg.lookup_column_name]
                ):
                    input_ids = input_ids[:, : data_cfg.max_length]
                    break

                truncate_length *= 2
            return {"input_ids": input_ids}

        dataset = (
            dataset.shuffle(self.quant_config.seed)
            .select(range(data_cfg.num_samples))
            .select_columns([data_cfg.lookup_column_name])
            .map(function=preprocess)
        )

        return dataset

    @torch.no_grad()
    def _perform_smoothing(
        self,
        activation_norms: List[torch.Tensor],
        fc_weights: List[torch.Tensor],
        activation_max: torch.Tensor,
        *,
        migration_strength: float = 0.5,
        epsilon: float = 1e-5,
        inplace: bool = False,
    ) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        """Perform activation-weight smoothing in SmoothQuant.

        Performs the activation-weight smoothing scheme described in SmoothQuant
        (Xiao et al., 2023), which migrates the amplitude of outliers from activations
        to weights of matmul layers. The function takes in the following parameters:

        Args:
            activation_norms: torch.Tensors representing affine parameters
                (i.e., beta and gamma) of a normalization layer before each matmul layer.
            fc_weights: torch.Tensors representing the weight matrices of the matmul layer.
            activation_max: The maximum activation value of inputs of the matmul layer.
            migration_strength: the strength of the activation migration. Default is 0.5.
            epsilon: The epsilon used for numerical stability when calculating the scales.
                Default is 1e-5.

        Returns:
            A tuple of three torch.Tensors: (smoothed_activation_norms, smoothed_fc_weights)

        The function calculates "scales" as `pow(|Activation|, migration_strength) /
        pow(|Weight|, 1-migration_strength)` and applies the smoothing effect into
        a normalization layer that exists before every matmul layer. This is done because
        it is more efficient than introducing a new smoothing layer before every matmul layer.
        Fusing the smoothing effect into the normalization layer results in a faster and
        more efficient implementation of the smoothing scheme.

        The function returns the smoothed normalization coefficients and the smoothed weight
        matrices after the smoothing process.
        """
        # shape of activation norms: [InChannels]
        # shape of fc weights: [OutChannels, InChannels]
        # shape of activation_max: [InChannels]

        # pylint: disable=too-many-locals
        assert activation_norms
        assert fc_weights

        assert activation_norms[0].ndim == 1
        in_channels = activation_norms[0].size(0)
        device = activation_norms[0].device
        dtype = activation_norms[0].dtype

        for norm in activation_norms:
            assert tuple(norm.size()) == (in_channels,)
            assert norm.device == device
            assert norm.dtype == dtype

        for weight in fc_weights:
            assert weight.ndim == 2
            assert weight.size(1) == in_channels
            assert weight.device == device
            assert weight.dtype == dtype

        activation_max = activation_max.to(device=device)
        weight_max = fc_weights[0].abs().max(dim=0).values
        for weight in fc_weights[1:]:
            weight_max = torch.maximum(weight_max, weight.abs().max(dim=0).values)

        assert tuple(activation_max.size()) == (in_channels,)
        assert tuple(weight_max.size()) == (in_channels,)
        alpha = migration_strength
        scales = (
            (
                activation_max.to(dtype=torch.float32).pow(alpha)
                / weight_max.to(dtype=torch.float32).pow(1 - alpha)
            )
            .clamp(min=epsilon)
            .to(dtype=dtype)
        )

        scaled_activation_norms = [act_norm / scales for act_norm in activation_norms]
        scaled_weights = [w * scales.view(1, -1) for w in fc_weights]

        if inplace:
            for dst, src in zip(activation_norms, scaled_activation_norms):
                dst.copy_(src)
            for dst, src in zip(fc_weights, scaled_weights):
                dst.copy_(src)

        return scaled_activation_norms, scaled_weights

    def _smooth(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Smooths the models before Quantization."""
        model.to(device=torch.device(self.quant_config.device))
        model.eval()
        model = cast(SmoothQuantHook, self.hook).pre_smooth(model)

        # collect stats for SmoothQuant scale.
        dataset = self.get_calib_dataset()
        quant_config = cast(SmoothQuantConfig, self.quant_config)
        max_input_stats, _ = collect_stats(
            model,
            quant_config.device,
            dataset,
            cast(SmoothQuantHook, self.hook).get_linear_layer_types(),
            tqdm_desc="Collecting stats for Smoothing.",
            percentile=100.0,
        )

        # TODO change name to pre_act_params, post_act_params
        # (attn_fc, ff2 are not scaled with norms)
        for norms, weights, name in cast(
            SmoothQuantHook, self.hook
        ).iter_smooth_norm_weights(model):
            self._perform_smoothing(
                norms,
                weights,
                max_input_stats[name],
                migration_strength=quant_config.smoothquant_args.migration_strength,
                inplace=True,
            )

    def pre_quantize(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Pre-procedure that should be called before quantize() is called."""
        self._smooth(model)

    def quantize(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Quantize model with SmoothQuant."""
        dataset = self.get_calib_dataset()
        max_input_stats, max_output_stats = collect_stats(
            model,
            self.quant_config.device,
            dataset,
            cast(SmoothQuantHook, self.hook).get_linear_layer_types(),
            percentile=self.quant_config.percentile,
            tqdm_desc="Collecting stats for Static Quantization.",
        )
        for quant_input in self.hook.iter_tf_quant_inputs(model):
            quant_result = cast(SmoothQuantHook, self.hook).get_quant_result(
                quant_input,
                max_input_stats=max_input_stats,
                max_output_stats=max_output_stats,
            )

            for field in fields(quant_result):
                layer_quant_result = getattr(quant_result, field.name)
                if isinstance(layer_quant_result, WeightActQuantResult):
                    layer = model.get_submodule(layer_quant_result.module_name)
                    q_layer = WeightActQuantizedLinearLayer.from_layer(
                        layer, layer_quant_result
                    )
                    quant_result.block.add_module(field.name, q_layer)

        return model



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Savers to save a converted checkpoints into various file types."""

from __future__ import annotations

import os
from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import Dict, Union

import h5py  # type: ignore[import]
import numpy as np
import safetensors.numpy  # type: ignore[import]
from typing_extensions import Self

from friendli.enums import CheckpointFileType
from friendli.errors import CheckpointConversionError
from friendli.logging import logger


def get_saver(
    ckpt_file_type: CheckpointFileType, output_path: Union[str, os.PathLike]
) -> CheckpointSaver:
    """Create a saver that corresponds to the file type."""
    if ckpt_file_type == CheckpointFileType.HDF5:
        return HDF5Saver(output_path)
    if ckpt_file_type == CheckpointFileType.SAFETENSORS:
        return SafetensorsSaver(output_path)
    raise CheckpointConversionError(
        f"Output file type {ckpt_file_type} is not supported."
    )


class CheckpointSaver(AbstractContextManager):
    """Abstract for savers."""

    def __init__(self, output_path: Union[str, os.PathLike]) -> None:
        """Check that the output file already exists."""
        super().__init__()
        self._output_path = output_path

    @abstractmethod
    def save_tensor(self, tensor_id: str, t: np.ndarray) -> None:
        """Save the tensor in the file."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the output checkpoint file."""
        raise NotImplementedError

    def __enter__(self) -> Self:
        """Enter for context manager."""
        return self

    def __exit__(self, *exc) -> None:
        """Exit for context manager."""
        self.close()


class HDF5Saver(CheckpointSaver):
    """Saver for HDF5."""

    def __init__(self, output_path: Union[str, os.PathLike]) -> None:
        """Create a HDF5 file."""
        super().__init__(output_path)
        self._out_f = h5py.File(output_path, "w")

    def save_tensor(self, tensor_id: str, t: np.ndarray) -> None:
        """Create a group if not exists, and save the tensor in the file."""
        self._out_f[tensor_id] = t

    def close(self) -> None:
        """Close the HDF5 file."""
        self._out_f.close()


class SafetensorsSaver(CheckpointSaver):
    """Saver for Safetensors.

    This temporally saves the converted tensors in local memory.
    Then, all of the tensors are saved in the file at a time when close() is called,
    because Safetensors does not support stream saving.
    """

    def __init__(self, output_path: Union[str, os.PathLike]) -> None:
        """Initialize a saver."""
        super().__init__(output_path)
        self._output_path = output_path
        self._tensors: Dict[str, np.ndarray] = {}

    def save_tensor(self, tensor_id: str, t: np.ndarray) -> None:
        """Save the tensor in the local memory."""
        self._tensors[tensor_id] = t

    def _save_to_file(self) -> None:
        """Save the tensors in the file."""
        logger.info("Saving the converted checkpoint...")
        safetensors.numpy.save_file(self._tensors, self._output_path)

    def close(self) -> None:
        """Save the tensors in the file."""
        self._save_to_file()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Team GQL Clients."""

from __future__ import annotations

from typing import Any, Dict, List

from friendli.client.graphql.base import GqlClient

TeamProjectsOp = """
query ClientTeam($input: ID!) {
  clientTeam(id: $input) {
    projects {
      edges {
        node {
          id
          name
          state
        }
      }
    }
  }
}
"""


class TeamGqlClient(GqlClient):
    """Team gql client."""

    def get_projects(self, team_id: str) -> List[Dict[str, Any]]:
        """List team projects."""
        response = self.run(query=TeamProjectsOp, variables={"input": team_id})
        return response["clientTeam"]["projects"]["edges"]

    def get_project_ids(self, team_id: str) -> List[str]:
        """List project IDs."""
        response = self.run(query=TeamProjectsOp, variables={"input": team_id})
        return [
            edge["node"]["id"] for edge in response["clientTeam"]["projects"]["edges"]
        ]



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Optional as _Optional

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message

DESCRIPTOR: _descriptor.FileDescriptor

class V1TextToImageRequest(_message.Message):
    __slots__ = [
        "prompt",
        "negative_prompt",
        "num_outputs",
        "num_inference_steps",
        "guidance_scale",
        "seed",
        "response_format",
    ]
    PROMPT_FIELD_NUMBER: _ClassVar[int]
    NEGATIVE_PROMPT_FIELD_NUMBER: _ClassVar[int]
    NUM_OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    NUM_INFERENCE_STEPS_FIELD_NUMBER: _ClassVar[int]
    GUIDANCE_SCALE_FIELD_NUMBER: _ClassVar[int]
    SEED_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FORMAT_FIELD_NUMBER: _ClassVar[int]
    prompt: str
    negative_prompt: str
    num_outputs: int
    num_inference_steps: int
    guidance_scale: float
    seed: int
    response_format: str
    def __init__(
        self,
        prompt: _Optional[str] = ...,
        negative_prompt: _Optional[str] = ...,
        num_outputs: _Optional[int] = ...,
        num_inference_steps: _Optional[int] = ...,
        guidance_scale: _Optional[float] = ...,
        seed: _Optional[int] = ...,
        response_format: _Optional[str] = ...,
    ) -> None: ...



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chat_completions.proto
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16\x63hat_completions.proto"\xd3\x03\n\x18V1ChatCompletionsRequest\x12\x33\n\x08messages\x18\x01 \x03(\x0b\x32!.V1ChatCompletionsRequest.Message\x12\x1e\n\x11\x66requency_penalty\x18\x03 \x01(\x02H\x00\x88\x01\x01\x12\x17\n\nmax_tokens\x18\x05 \x01(\x05H\x01\x88\x01\x01\x12\x0e\n\x01n\x18\x06 \x01(\x05H\x02\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x07 \x01(\x02H\x03\x88\x01\x01\x12\x0c\n\x04stop\x18\x08 \x03(\t\x12\x13\n\x06stream\x18\t \x01(\x08H\x04\x88\x01\x01\x12\x18\n\x0btemperature\x18\n \x01(\x02H\x05\x88\x01\x01\x12\x12\n\x05top_p\x18\x0b \x01(\x02H\x06\x88\x01\x01\x12!\n\x14timeout_microseconds\x18\x1e \x01(\x05H\x07\x88\x01\x01\x1a(\n\x07Message\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\t\x12\x0c\n\x04role\x18\x02 \x01(\tB\x14\n\x12_frequency_penaltyB\r\n\x0b_max_tokensB\x04\n\x02_nB\x13\n\x11_presence_penaltyB\t\n\x07_streamB\x0e\n\x0c_temperatureB\x08\n\x06_top_pB\x17\n\x15_timeout_microsecondsb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chat_completions_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_V1CHATCOMPLETIONSREQUEST"]._serialized_start = 27
    _globals["_V1CHATCOMPLETIONSREQUEST"]._serialized_end = 494
    _globals["_V1CHATCOMPLETIONSREQUEST_MESSAGE"]._serialized_start = 328
    _globals["_V1CHATCOMPLETIONSREQUEST_MESSAGE"]._serialized_end = 368
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Text to Image API."""

# pylint: disable=line-too-long, no-name-in-module

from __future__ import annotations

import os
from typing import Optional, Type

from friendli.schema.api.v1.codegen.text_to_image_pb2 import V1TextToImageRequest
from friendli.schema.api.v1.images.image import Image, ImageResponseFormatParam
from friendli.sdk.api.base import AsyncServingAPI, ServingAPI
from friendli.utils.compat import model_parse


class TextToImage(ServingAPI[Type[V1TextToImageRequest]]):
    """Text to image API."""

    @property
    def _api_path(self) -> str:
        return "v1/text-to-image"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        boundary = os.urandom(16).hex().encode("ascii")
        return f"multipart/form-data; boundary={boundary.decode('ascii')}"

    @property
    def _request_pb_cls(self) -> Type[V1TextToImageRequest]:
        return V1TextToImageRequest

    def create(
        self,
        *,
        prompt: str,
        model: Optional[str] = None,
        negative_prompt: Optional[str] = None,
        num_outputs: Optional[int] = None,
        num_inference_steps: Optional[int] = None,
        guidance_scale: Optional[float] = None,
        seed: Optional[int] = None,
        response_format: Optional[ImageResponseFormatParam] = None,
    ) -> Image:
        """Creates an image from the text prompt.

        Args:
            prompt (str): Text prompt describing the image to generate.
            model (Optional[str], optional): Code of the model to use.
            negative_prompt (Optional[str], optional): A text specifying what you don't want in your image(s). Defaults to None.
            num_outputs (Optional[int], optional): The number of images to generate. Must be between 1 and 16. Only 1 output will be generated when not provided. Defaults to None.
            num_inference_steps (Optional[int], optional): The number of inference steps for denoising process. 50 steps will be taken when not provided. Defaults to None.
            guidance_scale (Optional[float], optional): Guidance scale to control how much generation process adheres to the text prompt. When not provided, it is set to 7.5. Defaults to None.
            seed (Optional[int], optional): Seed to control random procedure. If nothing is given, the API generate the seed randomly. Defaults to None.
            response_format (Optional[ImageResponseFormatParam], optional): The format in which the generated images are returned. Defaults to None.

        Returns:
            Image: Data of the generated image.

        """
        request_dict = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "num_outputs": num_outputs,
            "num_inference_steps": num_inference_steps,
            "guidance_scale": guidance_scale,
            "seed": seed,
            "response_format": response_format,
        }
        response = self._request(data=request_dict, stream=False, model=model)

        return model_parse(Image, response.json())


class AsyncTextToImage(AsyncServingAPI[Type[V1TextToImageRequest]]):
    """Text to image API."""

    @property
    def _api_path(self) -> str:
        return "v1/text-to-image"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        boundary = os.urandom(16).hex().encode("ascii")
        return f"multipart/form-data; boundary={boundary.decode('ascii')}"

    @property
    def _request_pb_cls(self) -> Type[V1TextToImageRequest]:
        return V1TextToImageRequest

    async def create(
        self,
        *,
        prompt: str,
        model: Optional[str] = None,
        negative_prompt: Optional[str] = None,
        num_outputs: Optional[int] = None,
        num_inference_steps: Optional[int] = None,
        guidance_scale: Optional[float] = None,
        seed: Optional[int] = None,
        response_format: Optional[ImageResponseFormatParam] = None,
    ) -> Image:
        """Creates an image from the text prompt asychronously.

        Args:
            prompt (str): Text prompt describing the image to generate.
            model (Optional[str], optional): Code of the model to use.
            negative_prompt (Optional[str], optional): A text specifying what you don't want in your image(s). Defaults to None.
            num_outputs (Optional[int], optional): The number of images to generate. Must be between 1 and 16. Only 1 output will be generated when not provided. Defaults to None.
            num_inference_steps (Optional[int], optional): The number of inference steps for denoising process. 50 steps will be taken when not provided. Defaults to None.
            guidance_scale (Optional[float], optional): Guidance scale to control how much generation process adheres to the text prompt. When not provided, it is set to 7.5. Defaults to None.
            seed (Optional[int], optional): Seed to control random procedure. If nothing is given, the API generate the seed randomly. Defaults to None.
            response_format (Optional[ImageResponseFormatParam], optional): The format in which the generated images are returned. Defaults to None.

        Returns:
            Image: Data of the generated image.

        """
        request_dict = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "num_outputs": num_outputs,
            "num_inference_steps": num_inference_steps,
            "guidance_scale": guidance_scale,
            "seed": seed,
            "response_format": response_format,
        }
        response = await self._request(data=request_dict, stream=False, model=model)

        return model_parse(Image, response.json())



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Generates with Friendli Serverless Endpoints APIs."""

from __future__ import annotations

import typer

from friendli.cli.api import chat_completions, completions, text_to_image

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)

app.add_typer(chat_completions.app, name="chat-completions", help="Chat completions.")
app.add_typer(completions.app, name="completions", help="Text completions.")
app.add_typer(text_to_image.app, name="text-to-image", help="Text-to-image.")



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Compatibility Utils."""

from __future__ import annotations

from typing import Any, Dict, Type, TypeVar, cast

import pydantic

_ModelT = TypeVar("_ModelT", bound=pydantic.BaseModel)

PYDANTIC_V2 = pydantic.VERSION.startswith("2.")


def model_parse(model: Type[_ModelT], data: Any) -> _ModelT:
    """Parse a pydantic model from data."""
    if PYDANTIC_V2:
        return model.model_validate(data)  # type: ignore
    return model.parse_obj(data)  # type: ignore


def model_dump(
    model: pydantic.BaseModel,
    *,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
) -> dict[str, Any]:
    """Dump data from a pydantic model."""
    if PYDANTIC_V2:
        return model.model_dump(  # type: ignore
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
    return cast(
        Dict[str, Any],
        model.dict(  # type: ignore
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        ),
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""V1 Checkpoint Attributes Schemas."""

from __future__ import annotations

from typing import Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Extra, Field
from typing_extensions import Annotated

from friendli.enums import ModelDataType, QuantMode
from friendli.utils.compat import PYDANTIC_V2


class V1CommonAttributes(BaseModel):
    """V1 checkpoint attributes schema."""

    if PYDANTIC_V2:
        model_config = ConfigDict(protected_namespaces=(), extra=Extra.forbid)  # type: ignore
    else:
        model_config = ConfigDict(extra=Extra.forbid)

    dtype: ModelDataType
    quant_scheme: Optional[QuantMode] = None
    quant_group_size: Optional[int] = None
    quant_bit: Optional[int] = None


class V1BlenderbotAttributes(V1CommonAttributes):
    """V1 Blenderbot attributes schema."""

    model_type: Literal["blenderbot"]
    head_size: int
    num_heads: int
    hidden_size: int
    ff_intermediate_size: int
    num_encoder_layers: int
    num_decoder_layers: int
    max_input_length: int
    max_output_length: int
    vocab_size: int
    eos_token: int
    decoder_start_token: int


class V1BloomAttributes(V1CommonAttributes):
    """V1 Bloom attributes schema."""

    model_type: Literal["bloom"]
    head_size: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int


class V1GPTAttributes(V1CommonAttributes):
    """V1 GPT attributes schema."""

    model_type: Literal["gpt"]
    head_size: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int


class V1GPTJAttributes(V1CommonAttributes):
    """V1 GPT-J attributes schema."""

    model_type: Literal["gpt-j"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int
    rope_theta: Optional[float]


class V1GPTNeoXAttributes(V1CommonAttributes):
    """V1 GPT-NeoX attributes schema."""

    model_type: Literal["gpt-neox"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int
    rope_theta: Optional[float]


class V1GPTNeoXHFAttributes(V1CommonAttributes):
    """V1 GPT-NeoX HF attributes schema."""

    model_type: Literal["gpt-neox-hf"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int
    rope_theta: Optional[float]


class V1LlamaAttributes(V1CommonAttributes):
    """V1 LLaMA attributes schema."""

    model_type: Literal["llama"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_kv_heads: int
    num_layers: int
    ff_intermediate_size: int
    max_length: int
    vocab_size: int
    eos_token: int
    rope_theta: Optional[float]


class V1MistralAttributes(V1CommonAttributes):
    """V1 Mistral attributes schema."""

    model_type: Literal["mistral"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_kv_heads: int
    num_layers: int
    ff_intermediate_size: int
    max_length: int
    vocab_size: int
    eos_token: int
    attention_window_size: int


class V1OPTAttributes(V1CommonAttributes):
    """V1 OPT attributes schema."""

    model_type: Literal["opt"]
    head_size: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int


class V1T5Attributes(V1CommonAttributes):
    """V1 T5 attributes schema."""

    model_type: Literal["t5", "t5-v1_1"]
    head_size: int
    num_heads: int
    hidden_size: int
    ff_intermediate_size: int
    num_encoder_layers: int
    num_decoder_layers: int
    max_input_length: int
    max_output_length: int
    num_pos_emb_buckets: int
    max_pos_distance: int
    vocab_size: int
    eos_token: int
    decoder_start_token: int


class V1FalconAttributes(V1CommonAttributes):
    """V1 Falcon attributes schema."""

    model_type: Literal["falcon-7b", "falcon"]
    head_size: int
    num_heads: int
    num_kv_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int
    rotary_dim: int
    rope_theta: Optional[float]


class V1MPTAttributes(V1CommonAttributes):
    """V1 MPT attributes schema."""

    model_type: Literal["mpt"]
    head_size: int
    num_heads: int
    num_layers: int
    max_length: int
    vocab_size: int
    eos_token: int
    clip_qkv: float
    num_kv_heads: Optional[int]


class V1PhiAttributes(V1CommonAttributes):
    """V1 Phi attributes schema."""

    model_type: Literal["phi"]
    head_size: int
    rotary_dim: int
    num_heads: int
    num_kv_heads: int
    num_layers: int
    ff_intermediate_size: int
    max_length: int
    vocab_size: int
    eos_token: int
    rope_theta: Optional[float]


V1CheckpointAttributes = Annotated[
    Union[
        V1BlenderbotAttributes,
        V1BloomAttributes,
        V1FalconAttributes,
        V1GPTAttributes,
        V1GPTJAttributes,
        V1GPTNeoXAttributes,
        V1GPTNeoXHFAttributes,
        V1LlamaAttributes,
        V1MistralAttributes,
        V1MPTAttributes,
        V1OPTAttributes,
        V1T5Attributes,
        V1PhiAttributes,
    ],
    Field(discriminator="model_type"),
]


class V1AttributesValidationModel(BaseModel):
    """Model for validating attributes."""

    model_config = ConfigDict(extra=Extra.forbid)

    attr: V1CheckpointAttributes



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Converter Interface."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Any, Dict, List, Tuple

import numpy as np
import torch
from tqdm import tqdm

from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
)


class ModelConversionInterface(ABC):
    """Interface get information for converting models."""

    @abstractmethod
    def get_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Get list of conversion informations for the model."""

    @abstractmethod
    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""

    @abstractmethod
    def check_config(self) -> None:
        """Check if the model is convertable."""

    def convert(
        self,
        model: torch.nn.Module,
        convert_info_list: List[ConvertInfo],
    ) -> Generator[Tuple[str, np.ndarray], None, None]:
        """Convert Huggingface Model to Friendli format(.h5).

        Args:
            model (torch.nn.Module): Huggingface model.
            output_path (str): Path to save the converted checkpoint.
            convert_info_list (List[ConvertInfo]):
                List of convert information of the parameter in huggingface checkpoint.
        """
        state_dict = model.state_dict()
        total_layers = len(convert_info_list)
        with tqdm(total=total_layers, desc="Converting", unit="tensor") as pbar:
            for convert_info in convert_info_list:
                converted_name, reshape_fn, param_names, data_type = (
                    convert_info.converted_name,
                    convert_info.reshape_fn,
                    convert_info.param_names,
                    convert_info.data_type,
                )
                params = [
                    get_tensor_from_state_dict(state_dict, param_name)
                    for param_name in param_names
                ]
                reshaped_tensor = reshape_fn(params)
                yield (
                    converted_name,
                    convert_tensor_to_np_array(reshaped_tensor, data_type),
                )
                pbar.update()


class NonTFBlockConversionInterface(ABC):
    """Interface get information for converting common layers."""

    @property
    @abstractmethod
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for the non-transformer blocks."""


class DecoderTFBlockConversionInterface(ABC):
    """Interface get information for converting decoder layers."""

    @property
    @abstractmethod
    def decoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the decoder's transformer block number."""

    @property
    @abstractmethod
    def decoder_layer_num(self) -> int:
        """Return the number of transformer blocks in the decoder."""

    @property
    @abstractmethod
    def decoder_hidden_size(self) -> int:
        """Return the hidden size of the decoder."""

    @property
    @abstractmethod
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads."""

    @property
    @abstractmethod
    def decoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the decoder."""

    @property
    @abstractmethod
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for transformer blocks in the decoder."""

    @property
    @abstractmethod
    def decoder_head_size(self) -> int:
        """Return the head size of the decoder."""

    @property
    @abstractmethod
    def decoder_ff_intermediate_size(self) -> int:
        """Return the intermediate size of the linear layer in decoder's MLP."""


class EncoderTFBlockConversionInterface(ABC):
    """Interface get information for converting encoder layers."""

    @property
    @abstractmethod
    def encoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the encoder's transformer block number."""

    @property
    @abstractmethod
    def encoder_layer_num(self) -> int:
        """Return the number of transformer blocks in the encoder."""

    @property
    @abstractmethod
    def encoder_hidden_size(self) -> int:
        """Return the hidden size of the encoder."""

    @property
    @abstractmethod
    def encoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the encoder."""

    @property
    @abstractmethod
    def encoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for transformer blocks in the encoder."""

    @property
    @abstractmethod
    def encoder_head_size(self) -> int:
        """Return the head size of the encoder."""

    @property
    @abstractmethod
    def encoder_ff_intermediate_size(self) -> int:
        """Return the intermediate size of the linear layer in encoder's MLP."""


class RotaryEmbeddingConversionInterface(ABC):
    """Interface get information for converting rotary embeddings."""

    @property
    @abstractmethod
    def rotary_dim(self) -> int:
        """Return the dimension of rotary embeddings."""

    @property
    @abstractmethod
    def rotary_emb_base(self) -> float:
        """Return the base of rotary embeddings."""



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI Validation Utilities."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from importlib.util import find_spec
from typing import Any, Dict, Optional, Type

import typer
from pydantic import ValidationError

from friendli.errors import InvalidAttributesError, InvalidConfigError
from friendli.schema.resource.v1.attributes import V1AttributesValidationModel
from friendli.utils.compat import model_parse
from friendli.utils.version import (
    FRIENDLI_PACKAGE_NAME,
    get_installed_version,
    get_latest_version,
    is_latest_version,
)


def validate_datetime_format(datetime_str: Optional[str]) -> Optional[str]:
    """Validate the datetime format."""
    if datetime_str is None:
        return None

    try:
        datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S")
    except ValueError as exc:
        raise typer.BadParameter(
            "The datetime format should be {YYYY}-{MM}-{DD}T{HH}:{MM}:{SS}"
        ) from exc
    return datetime_str


def check_package_version() -> None:
    """Validate the installed CLI version."""
    installed_version = get_installed_version()
    if not is_latest_version(installed_version):
        latest_version = get_latest_version()
        typer.secho(
            f"Package version({installed_version}) is not the latest. "
            f"We recommend installing the latest version({latest_version}) with "
            f"'pip install {FRIENDLI_PACKAGE_NAME}=={latest_version} -U'.",
            fg=typer.colors.YELLOW,
        )


def validate_checkpoint_attributes(attr: Dict[str, Any]) -> None:
    """Validate checkpoint attributes schema."""
    try:
        model_parse(V1AttributesValidationModel, {"attr": attr})
    except ValidationError as exc:
        msgs = []
        for error in exc.errors():
            error_type = error["type"]
            if error_type == "union_tag_invalid":
                msgs.append("'model_type' is not one of the expected values.")
            elif error_type == "union_tag_not_found":
                msgs.append("'model_type' filed is missing.")
            else:
                msgs.append(f"{error['msg']}. Correct the field '{error['loc'][-1]}'")

        if len(msgs) == 1:
            msg = msgs[0]
        else:
            msg = "\n>>> " + "\n>>> ".join(msgs)
        raise InvalidAttributesError(msg) from exc


def validate_enums(val: Any, enum_cls: Type[Enum]) -> Any:
    """Validate if the value is the proper enum."""
    try:
        return enum_cls(val)
    except ValueError as exc:
        supported_values = set([e.value for e in enum_cls])
        raise InvalidConfigError(
            f"Invalid value. Please provide one of {supported_values}"
        ) from exc


def validate_convert_imports() -> None:
    """Validate the import modules for checkpoint conversion."""
    if find_spec("torch") is None:
        raise ModuleNotFoundError(
            "To convert the checkpoint, you must install 'torch'."
        )
    if find_spec("transformers") is None or find_spec("accelerate") is None:
        raise ModuleNotFoundError(
            "To convert the checkpoint,"
            " your must install the package with 'pip install \"friendli-client[mllib]\"'"
        )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Converter Utils."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import numpy as np
import torch
from peft import PeftConfig  # type: ignore[import] # pylint: disable=import-error
from transformers import (  # type: ignore[import]
    AutoConfig,
    AutoTokenizer,
    GenerationConfig,
    PretrainedConfig,
)

from friendli.enums import ModelDataType
from friendli.errors import (
    CheckpointConversionError,
    NotFoundError,
    NotSupportedCheckpointError,
    TokenizerNotFoundError,
)
from friendli.logging import logger


def convert_to_gpt_j_params(param: torch.Tensor, rotary_dim: int) -> torch.Tensor:
    """Reshape weight or bias tensor with rotary embedding to gpt-j format.

    Args:
        param (torch.Tensor): Target tensor to convert. Shape must be (num_heads, head_size, ...)
        rotary_dim (int): Degree of rotary embedding

    Returns:
        Torch tensor that heads are rotated.

    Raises:
        CheckpointConversionError: If arguments do not satisfy the requirements.

    """
    if param.ndim < 2:
        raise CheckpointConversionError(
            "Tensor dimension should be greater or equal than 2 for rotary conversion, "
            f"but got {param.ndim}"
        )

    head_size = param.shape[1]
    if rotary_dim > head_size:
        raise CheckpointConversionError(
            f"'rotary_dim' ({rotary_dim}) should be less or equal than 'head_size' ({head_size})"
        )

    param_rot = param[:, :rotary_dim]
    param_pass = param[:, rotary_dim:]

    origin_shape = param_rot.shape
    param_rot_1 = param_rot[:, : rotary_dim // 2]
    param_rot_2 = param_rot[:, rotary_dim // 2 :]
    param_rot = torch.stack((param_rot_1, param_rot_2), dim=2).reshape(*origin_shape)

    return torch.cat((param_rot, param_pass), dim=1)


def get_tensor_from_state_dict(
    state_dict: Dict[str, Any], tensor_name: str
) -> torch.Tensor:
    """Get the tensor whose name is 'tensor_name' from 'state_dict'.

    Args:
        state_dict (Dict[str, Any]): Model checkpoint's state_dict.
        tensor_name (str): Name of tensor to get.

    Returns:
        Corresponding torch Tensor.

    Raises:
        CheckpointConversionError: If 'tensor_name' does not exist in 'state_dict'

    """
    if tensor_name not in state_dict:
        raise CheckpointConversionError(
            f"Cannot find '{tensor_name}' in the model checkpoint"
        )

    return state_dict[tensor_name]


def convert_tensor_to_np_array(
    param: torch.Tensor,
    data_type: ModelDataType,
) -> np.ndarray:
    """Reshape tensor to numpy ndarray.

    Args:
        param (torch.Tensor): The tensor to be converted.
        data_type (CheckpointDataType): The data type of the tensor.

    Returns:
        np.ndarray: The converted numpy ndarray from the tensor.

    """
    dtype_map = {
        ModelDataType.BF16: torch.bfloat16,
        ModelDataType.FP16: torch.float16,
        ModelDataType.FP32: torch.float32,
        ModelDataType.INT8: torch.int8,
        ModelDataType.INT4: torch.int8,
    }

    dtype = dtype_map[data_type]

    if data_type is ModelDataType.BF16:
        return (
            param.cpu()
            .detach()
            .to(dtype)
            .view(dtype=torch.float16)
            .numpy()
            .view(np.uint16)
        )

    if data_type is ModelDataType.INT4:
        pack_num = 8 // 4
        int4_param = torch.zeros(
            (param.shape[0], param.shape[1] // pack_num),
            dtype=torch.uint8,
            device=param.device,
        )
        for col in range(int4_param.shape[1]):
            for i in range(pack_num):
                int4_param[:, col] |= param[:, col * pack_num + i] << (i * 4)
        param = int4_param

    return param.to("cpu").detach().to(dtype).numpy()


def get_tokenizer(
    model_name_or_path: str,
    *,
    cache_dir: Optional[str] = None,
) -> AutoTokenizer:
    """Try to get tokenizer of a pretrained model."""
    try:
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            cache_dir=cache_dir,
            trust_remote_code=True,
        )
    except OSError as exc:
        raise TokenizerNotFoundError(str(exc)) from exc

    if not tokenizer.is_fast:
        raise TokenizerNotFoundError(
            "This model does not support Friendli-compatible tokenizer"
        )

    return tokenizer


def save_tokenizer(
    model_name_or_path: str,
    *,
    cache_dir: Optional[str] = None,
    save_dir: str,
) -> None:
    """Try to save `tokenizer.json` of a pretrained model."""
    if not os.path.isdir(save_dir):
        raise NotFoundError(f"Directory '{save_dir}' is not found.")

    tokenizer = get_tokenizer(model_name_or_path, cache_dir=cache_dir)
    saved_file_paths = tokenizer.save_pretrained(save_directory=save_dir)

    tokenizer_json_path = None
    for path in saved_file_paths:
        if "tokenizer.json" == os.path.basename(path):
            tokenizer_json_path = path
        else:
            # Remove unnecessary files.
            try:
                os.remove(path)
            except FileNotFoundError:
                logger.warn(
                    "Tried to delete unnecessary tokenizer file %s but the file "
                    "is not found.",
                    path,
                )

    if tokenizer_json_path is None:
        raise TokenizerNotFoundError(
            "This model has the Friendli-compatible tokenizer implementation, but "
            "'tokenizer.json' file is not found."
        )


def get_model_generation_config(
    model_name_or_path: str, cache_dir: Optional[str] = None
) -> Optional[GenerationConfig]:
    """Get HuggingFace model generation config."""
    try:
        generation_config = GenerationConfig.from_pretrained(
            model_name_or_path, cache_dir=cache_dir, trust_remote_code=True
        )
    except (OSError, TypeError):
        generation_config = None

    return generation_config


def get_model_pretrained_config(
    model_name_or_path: str, model_output_path: str, cache_dir: Optional[str] = None
) -> PretrainedConfig:
    """Get HuggingFace model configs."""
    try:
        config = AutoConfig.from_pretrained(
            model_name_or_path, cache_dir=cache_dir, trust_remote_code=True
        )
    except OSError as exc:  # from AutoConfig.from_pretrained()
        config_dir = Path(model_name_or_path)
        model_output_dir = Path(model_output_path).parent
        if config_dir.exists() and model_output_dir.absolute() == config_dir.absolute():
            raise NotFoundError(
                f"'output_dir' ({model_output_dir.as_posix()}) and "
                f"'model_name_or_path' ({model_name_or_path}) are the same. "
                "In such a case, checkpoints should be prepared in 'output_dir'."
            ) from exc
        raise NotFoundError(str(exc)) from exc

    return config


def get_model_arch(config: PretrainedConfig) -> str:
    """Get HuggingFace model architecture from config."""
    model_arch_list = cast(List[str], cast(PretrainedConfig, config).architectures)
    if len(model_arch_list) == 0:
        raise NotSupportedCheckpointError(
            invalid_option=f"'architectures={model_arch_list}'",
            valid_options=["non empty list of architectures"],
        )
    model_arch = model_arch_list[0]
    return model_arch


def get_adapter_config(
    adapter_name_or_path: str, cache_dir: Optional[str]
) -> PeftConfig:
    """Get PeftConfig for Adapter."""
    try:
        adapter_config = PeftConfig.from_pretrained(
            adapter_name_or_path, cache_dir=cache_dir, trust_remote_code=True
        )
    except ValueError as exc:
        raise NotFoundError(str(exc)) from exc
    return adapter_config



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Converter Schema."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List

import torch

from friendli.enums import ModelDataType


@dataclass
class ConvertInfo:
    """Dataclass for convert information of the parameter in huggingface checkpoint.

    Args:
        param_names(List[str]): List of parameter names in the huggingface checkpoint.
        data_type(CheckpointDataType): Data type of the parameter.
        converted_name(str): Name of the converted parameter.
        reshape_fn(Callable[[List[torch.tensor]], np.ndarray]):
            Function to reshape the tensor from the huggignface checkpoint.
    """

    param_names: List[str]
    data_type: ModelDataType
    converted_name: str
    reshape_fn: Callable[[List[torch.Tensor]], torch.Tensor]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Endpoint SDK."""

# pylint: disable=arguments-differ

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Type

from friendli.client.graphql.endpoint import EndpointGqlClient
from friendli.schema.resource.v1.endpoint import Endpoint
from friendli.sdk.resource.base import ResourceApi


@dataclass
class EndpointApi(ResourceApi[EndpointGqlClient, Endpoint, str]):
    """Endpoint resource API."""

    client: EndpointGqlClient

    @property
    def _resource_model(self) -> Type[Endpoint]:
        return Endpoint

    def create(
        self,
        *,
        name: str,
        model_repo: str,
        gpu_type: str,
        num_gpus: int,
    ) -> Endpoint:
        """Creates a new dedicated endpoint."""
        data = self.client.create(
            team_id=self._get_team_id(),
            project_id=self._get_project_id(),
            name=name,
            model_repo=model_repo,
            gpu_type=gpu_type,
            num_gpus=num_gpus,
        )
        endpoint = self._model_parse(data)

        return endpoint

    def get(self, eid: str, *args, **kwargs) -> Endpoint:
        """Get information of a endpoint."""
        data = self.client.get(endpoint_id=eid)
        endpoint = self._model_parse(data)
        return endpoint

    def list(self) -> List[Endpoint]:
        """List endpoints."""
        data = self.client.list(project_id=self._get_project_id())
        endpoints = self._model_parse(data)
        return endpoints

    def terminate(self, endpoint_id: str) -> None:
        """Terminate a endpoint."""
        self.client.terminate(
            team_id=self._get_team_id(),
            project_id=self._get_project_id(),
            endpoint_id=endpoint_id,
        )



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Mixtral Checkpoint Converter."""


from __future__ import annotations

from typing import Any, Dict, List, Optional, cast

from transformers import MixtralConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.models.llama import LlamaForCausalLMConverter
from friendli.modules.converter.schema import ConvertInfo


class MixtralForCausalLMConverter(LlamaForCausalLMConverter):
    """MixtralForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Mixtral architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if cast(MixtralConfig, self.config).hidden_act not in ["silu"]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'hidden_act={cast(MixtralConfig, self.config).hidden_act}'",
                    valid_options=["silu"],
                )
            if cast(MixtralConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(MixtralConfig, self.config).num_local_experts != 8:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'num_local_experts={cast(MixtralConfig, self.config).num_local_experts}",
                    valid_options=[8],
                )
            if cast(MixtralConfig, self.config).num_experts_per_tok != 2:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'num_experts_per_tok={cast(MixtralConfig, self.config).num_experts_per_tok}",
                    valid_options=[2],
                )

        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(MixtralConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The Mixtral model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "ff_intermediate_size": self.decoder_ff_intermediate_size,
            "max_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "rope_theta": self.rotary_emb_base,
            "num_experts": self.num_experts,
        }
        if isinstance(self.attention_window_size, int):
            # for sliding window
            attr["attention_window_size"] = self.attention_window_size
        return attr

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in LLaMA."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}input_layernorm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.weight",
                            f"{layer_prefix}self_attn.k_proj.weight",
                            f"{layer_prefix}self_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.o_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}post_attention_layernorm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}block_sparse_moe.gate.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}moe/router/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )
            for i in range(self.num_experts):
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}block_sparse_moe.experts.{i}.w1.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefix}moe/{i}/mlp/c_gate/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}block_sparse_moe.experts.{i}.w2.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefix}moe/{i}/mlp/c_proj/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}block_sparse_moe.experts.{i}.w3.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefix}moe/{i}/mlp/c_fc/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                    ]
                )
        return convert_info_list

    @property
    def model_type(self) -> str:
        """Model type."""
        return "mixtral"

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in Mixtral."""
        return cast(MixtralConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in Mixtral."""
        return cast(MixtralConfig, self.config).hidden_size

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of linear layer in Mixtral MoEs."""
        return cast(MixtralConfig, self.config).intermediate_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in Mixtral."""
        return cast(MixtralConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in Mixtral."""
        config = cast(MixtralConfig, self.config)
        if config.num_key_value_heads is None:
            return self.decoder_num_attention_heads
        return config.num_key_value_heads

    @property
    def attention_window_size(self) -> Optional[int]:
        """The size of sliding window attention in Mixtral."""
        return cast(MixtralConfig, self.config).sliding_window

    @property
    def num_experts(self) -> int:
        """The number of moe experts per transformer block in Mixtral."""
        return cast(MixtralConfig, self.config).num_local_experts

    @property
    def num_selected_moe_experts(self) -> int:
        """The number of selected moe experts per transformer block in Mixtral."""
        return cast(MixtralConfig, self.config).num_experts_per_tok



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Converter."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union, cast

import numpy as np
import torch
from peft import PeftType  # type: ignore[import] # pylint: disable=import-error
from peft.config import PeftConfig
from peft.tuners.lora import (  # type: ignore[import] # pylint: disable=import-error
    LoraConfig,
)
from transformers import GenerationConfig, PretrainedConfig  # type: ignore[import]

from friendli.enums import ModelDataType
from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.interface import (
    DecoderTFBlockConversionInterface,
    EncoderTFBlockConversionInterface,
    ModelConversionInterface,
    NonTFBlockConversionInterface,
)
from friendli.modules.converter.schema import ConvertInfo

SUPPORTED_GELU_FAMILY = [
    "gelu",
    "gelu_fast",
    "gelu_new",
    "gelu_python",
    "gelu_pytorch_tanh",
    "gelu_accurate",
]
SUPPORTED_HEAD_SIZE = [64, 80, 96, 128, 256]

MODEL_TYPE_TO_SUPPORTED_LORA_TARGET_MODULES_MAP = {
    "gptj": {"q_proj", "k_proj", "v_proj", "out_proj", "fc_in", "fc_out", "wte"},
    "llama": {
        "q_proj",
        "k_proj",
        "v_proj",
        "o_proj",
        "gate_proj",
        "up_proj",
        "down_proj",
    },
    "mpt": {"Wqkv", "out_proj", "up_proj", "down_proj"},
}
# TODO: remove this const map when engine supports lm head LoRA
MODEL_TYPE_TO_UNSUPPORTED_LORA_TARGET_MODULES_MAP = {
    "gptj": {"lm_head"},
    "llama": {"lm_head"},
    "mpt": {"lm_head"},
}

ENCODER_PREFIX = "encoder"
DECODER_PREFIX = "decoder"


class AbstractConverter(ModelConversionInterface, ABC):
    """Abstract class for converting Hugging Face checkpoint to Friendli checkpoint.

    Attributes:
        config (PreTrainedConfig): Hugging Face model configuration.
        generation_config (Optional[GenerationConfig]): Hugginface generation config.
            When set to None, `config` is used for configuring generation.
        data_type (CheckpointDataType): Data type for the Friendli checkpoint.
        quantize (bool): Whether to quantize the Friendli checkpoint.

    """

    def __init__(
        self,
        config: PretrainedConfig,
        generation_config: Optional[GenerationConfig],
        data_type: ModelDataType,
    ) -> None:
        """Initialize converter."""
        self.config = config
        self.generation_config = generation_config
        self.data_type = data_type

    def get_eos_token_id(self) -> Optional[int]:
        """Get ID of EOS token."""
        generation_eos_token_id = None
        if self.generation_config is not None:
            generation_eos_token_id = self.generation_config.eos_token_id

        config_eos_token_id = self.config.eos_token_id

        if generation_eos_token_id is None:
            eos_token_id = config_eos_token_id
        else:
            if generation_eos_token_id != config_eos_token_id:
                logger.warn(
                    "'eos_token' is different in generation_config (%s) and config (%s). "
                    "Please fill the correct value.",
                    generation_eos_token_id,
                    config_eos_token_id,
                )
                eos_token_id = None
            else:
                eos_token_id = config_eos_token_id

        if eos_token_id is None:
            logger.warn(
                "'eos_token' cannot be automatically configured. "
                "Please fill in the field by yourself."
            )

        return eos_token_id

    def token_embed_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape embedding layer's weight to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped embedding weight.

        """
        assert len(params) == 1
        return params[0]

    def pos_embed_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape position embedding layer's weight to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped position embedding weight.
        """
        assert len(params) == 1
        return params[0]

    def head_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape head layer's weight to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped head weight.

        """
        assert len(params) == 1
        return params[0]

    def linear_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape linear layer's weight to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped linear weight.

        """
        assert len(params) == 1
        param = params[0].transpose(0, 1)
        return param

    def linear_bias_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape linear layer's bias to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped linear bias.

        """
        assert len(params) == 1
        return params[0]

    def ln_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape layer norm layer's weight to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped layer norm weight.

        """
        assert len(params) == 1
        return params[0]

    def ln_bias_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape layer norm layer's bias to Friendli format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped layer norm bias.

        """
        assert len(params) == 1
        return params[0]

    def qkv_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape qkv layer's weight to Friendli format.

        In the original checkpoint, the qkv weight is stored as a single tensor or
        separated by three tensors. In the Friendli checkpoint, it is stored as a single tensor.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The tensor of reshaped qkv weight.

        """
        param = torch.cat(params, dim=0)
        param = param.transpose(0, 1)
        return param

    def qkv_bias_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """Reshape qkv layer's bias to Friendli format.

        In the original checkpoint, the qkv weight is stored as a single tensor or
        separated by three tensors. In the Friendli checkpoint, it is stored as a single tensor.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
           The tensor of reshaped qkv bias.

        """
        param = torch.cat(params, dim=0)
        return param


class DecoderOnlyConverter(
    AbstractConverter,
    NonTFBlockConversionInterface,
    DecoderTFBlockConversionInterface,
):
    """Converter for Decoder-Only models."""

    def check_config(self) -> None:
        """Check if a convertible form of the checkpoint from the decoder-only model config."""
        super().check_config()
        if self.decoder_head_size not in SUPPORTED_HEAD_SIZE:
            raise NotSupportedCheckpointError(
                invalid_option=f"decoder_head_size={self.decoder_head_size}",
                valid_options=SUPPORTED_HEAD_SIZE,
            )

    def get_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Get List of conversion informations for Decoder-Only model."""
        return self.non_transformer_convert_info_list + self.decoder_convert_info_list


class EncoderDecoderConverter(
    AbstractConverter,
    NonTFBlockConversionInterface,
    EncoderTFBlockConversionInterface,
    DecoderTFBlockConversionInterface,
):
    """Converter for Encoder-Decoder models."""

    def check_config(self) -> None:
        """Check if a convertible form of the checkpoint from the encoder-decoder model config."""
        if self.decoder_head_size not in SUPPORTED_HEAD_SIZE:
            raise NotSupportedCheckpointError(
                invalid_option=f"decoder_head_size={self.decoder_head_size}",
                valid_options=SUPPORTED_HEAD_SIZE,
            )

    def get_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Get list of conversion informations for Encoder-Decoder model."""
        return (
            self.non_transformer_convert_info_list
            + self.decoder_convert_info_list
            + self.encoder_convert_info_list
        )

    def get_decoder_start_token_id(self) -> Optional[int]:
        """Get ID of decoder start token."""
        generation_decoder_start_token_id = None
        if self.generation_config is not None:
            generation_decoder_start_token_id = (
                self.generation_config.decoder_start_token_id
            )

        config_decoder_start_token_id = self.config.decoder_start_token_id

        if generation_decoder_start_token_id is None:
            decoder_start_token_id = config_decoder_start_token_id
        else:
            if generation_decoder_start_token_id != config_decoder_start_token_id:
                logger.warn(
                    "'decoder_start_token_id' is different in generation_config "
                    "(%s) and config (%s). Please fill the correct value.",
                    generation_decoder_start_token_id,
                    config_decoder_start_token_id,
                )
                decoder_start_token_id = None
            else:
                decoder_start_token_id = config_decoder_start_token_id

        if decoder_start_token_id is None:
            logger.warn(
                "'decoder_start_token' cannot be automatically configured. "
                "Please fill in the field by yourself."
            )

        return decoder_start_token_id


class DecoderOnlyLoraConverter(AbstractConverter):
    """Converter for LoRA modules in the models."""

    def __init__(
        self,
        converter: AbstractConverter,
        adapter_config: PeftConfig,
    ) -> None:
        """Initialize LoRA Converter."""
        super().__init__(
            config=converter.config,
            generation_config=converter.generation_config,
            data_type=converter.data_type,
        )
        self.converter = cast(DecoderOnlyConverter, converter)
        self.adapter_config = cast(LoraConfig, adapter_config)

    def check_config(self) -> None:
        """Check if a convertible form of the checkpoint from the LoRAconfig."""
        if self.adapter_config.peft_type != PeftType.LORA:
            raise NotSupportedCheckpointError(
                invalid_option=f"peft_type={self.adapter_config.peft_type}",
                valid_options=[str(PeftType.LORA)],
            )
        if (
            self.config.model_type
            not in MODEL_TYPE_TO_SUPPORTED_LORA_TARGET_MODULES_MAP
        ):
            raise NotSupportedCheckpointError(
                invalid_option=f"model_type={self.config.model_type} for LORA",
                valid_options=list(
                    MODEL_TYPE_TO_SUPPORTED_LORA_TARGET_MODULES_MAP.keys()
                ),
            )
        if (
            self.adapter_config.layers_pattern is not None
            and len(self.adapter_config.layers_pattern) > 0
        ):
            raise NotSupportedCheckpointError(
                invalid_option=f"layers_pattern={self.adapter_config.layers_pattern}",
                valid_options=[None, [], ""],
            )
        if (
            self.adapter_config.rank_pattern is not None
            and len(self.adapter_config.rank_pattern) > 0
        ):
            raise NotSupportedCheckpointError(
                invalid_option=f"rank_pattern={self.adapter_config.rank_pattern}",
                valid_options=[None, {}],
            )
        if (
            self.adapter_config.alpha_pattern is not None
            and len(self.adapter_config.alpha_pattern) > 0
        ):
            raise NotSupportedCheckpointError(
                invalid_option=f"alpha_pattern={self.adapter_config.alpha_pattern}",
                valid_options=[None, {}],
            )

        if self.adapter_config.target_modules is not None:
            for target_module in self.adapter_config.target_modules:
                if (
                    target_module
                    not in MODEL_TYPE_TO_SUPPORTED_LORA_TARGET_MODULES_MAP[
                        self.config.model_type
                    ]
                ):
                    if (
                        target_module
                        in MODEL_TYPE_TO_UNSUPPORTED_LORA_TARGET_MODULES_MAP[
                            self.config.model_type
                        ]
                    ):
                        raise NotSupportedCheckpointError(
                            invalid_option=f"target_module={target_module}",
                            valid_options=list(
                                MODEL_TYPE_TO_SUPPORTED_LORA_TARGET_MODULES_MAP[
                                    self.config.model_type
                                ]
                            ),
                        )
                    logger.warn(
                        "Target module %s does not exist in the base model (%s). Will be ignored.",
                        target_module,
                        self.adapter_config.base_model_name_or_path,
                    )

        if (self.adapter_config.layers_to_transform is not None) and (
            self.adapter_config != list(range(self.converter.decoder_layer_num))
        ):
            raise NotSupportedCheckpointError(
                invalid_option=f"layers_to_transform={self.adapter_config.layers_to_transform}",
                valid_options=[
                    f"layers_to_transform=None"
                    f"layers_to_transform={list(range(self.converter.decoder_layer_num))}",
                ],
            )

    def get_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Get convert dict for LoRA model."""
        return self.adapter_convert_info_list

    def _get_layers_to_transform(self) -> List[int]:
        layers_to_transform = cast(LoraConfig, self.adapter_config).layers_to_transform
        if layers_to_transform is None:
            layers_to_transform = list(range(self.converter.decoder_layer_num))
        else:
            if isinstance(layers_to_transform, int):
                layers_to_transform = [layers_to_transform]
        return layers_to_transform

    def lora_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """Reshape LoRA layer's weight to Friendli format."""
        assert len(params) == 1
        return params[0].transpose(0, 1)

    def pre_convert(self, model: torch.nn.Module) -> torch.nn.Module:
        """Preprocess the adapter modules before converting.

        All the parameters of the LoRA low-rank matrixs are converted by `lora_weight_reshape`.
        If the parameter can't be converted by `lora_weight_reshape`,

        """
        return model

    def convert(  # pylint: disable=too-many-locals
        self,
        model: torch.nn.Module,
        convert_info_list: List[ConvertInfo],
    ) -> Generator[Tuple[str, np.ndarray], None, None]:
        """Reshape Lora adapter model's all layer to Friendli format."""
        model = self.pre_convert(model)
        yield from self.converter.convert(model, convert_info_list)

    def get_attributes(self) -> Dict[str, Any]:
        """Get adapter checkpoint attributes."""
        return {
            "name": "FILL ME",
            "type": "lora",
            "alpha": self.adapter_config.lora_alpha,
            "rank": self.adapter_config.r,
            "target-modules": list(self.adapter_target_modules),
            "ckpt-path": "FILL ME",
        }

    @property
    def adapter_target_modules(self) -> Set[str]:
        """Return the target modules that LoRA applies to."""
        if isinstance(self.adapter_config.target_modules, str):
            hf_target_modules = {self.adapter_config.target_modules}
        elif isinstance(self.adapter_config.target_modules, Iterable):
            hf_target_modules = set(self.adapter_config.target_modules)
        else:
            raise CheckpointConversionError("`target_modules` should not be None")

        translated_target_modules = set()
        for target in hf_target_modules:
            if target in self.adapter_target_module_map:
                translated_target_modules.add(self.adapter_target_module_map[target])

        return translated_target_modules

    @property
    @abstractmethod
    def adapter_target_module_map(self) -> Dict[str, str]:
        """Return the dictionary that maps Hugging Face's module name to Friendli's module name."""

    @property
    @abstractmethod
    def adapter_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for LoRA modules of the model."""


OneOfAdapterConverter = DecoderOnlyLoraConverter
OneOfConverter = Union[EncoderDecoderConverter, DecoderOnlyConverter]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli GPT2LMHeadModel QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List, Tuple, Type, cast

import torch
from transformers.models.gpt2 import GPT2LMHeadModel  # type: ignore[import]
from transformers.pytorch_utils import Conv1D  # type: ignore[import]

from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantGPT2Hook(SmoothQuantHook):
    """SmoothQuant Hook for GPT2LMHeadModel."""

    def iter_smooth_norm_weights(
        self, model: GPT2LMHeadModel
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in GPT2LMHeadModel."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.transformer.h):  # type: ignore[union-attr]
            # [LayerNorm 1] - [ QKV projection ] gets smoothed
            yield (
                [
                    decoder_layer.ln_1.weight.data,
                    decoder_layer.ln_1.bias.data,
                ],
                [
                    decoder_layer.attn.c_attn.weight.data.transpose(
                        0, 1
                    ),  # [OutDim, InDim]
                ],
                f"{self.quantized_layer_prefix}{index}.attn.c_attn",  # the input tensors fed into Q, K, V matrices are identical.
            )
            # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
            yield (
                [
                    decoder_layer.ln_2.weight.data,
                    decoder_layer.ln_2.bias.data,
                ],
                [decoder_layer.mlp.c_fc.weight.data.transpose(0, 1)],  # [OutDim, InDim]
                f"{self.quantized_layer_prefix}{index}.mlp.c_fc",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data.transpose(0, 1)],
                    [decoder_layer.attn.c_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.attn.c_proj",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data.transpose(0, 1)],
                    [decoder_layer.mlp.c_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.mlp.c_proj",
                )

    def iter_tf_quant_inputs(self, model: GPT2LMHeadModel) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of GPT2LMHeadModel."""
        for index, decoder_layer in enumerate(model.transformer.h):
            attn = decoder_layer.attn
            attn_weight_outdim = attn.c_attn.nf  # OutDim
            fc1 = decoder_layer.mlp.c_fc
            fc2 = decoder_layer.mlp.c_proj

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    attn.c_attn.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.attn.c_attn",
                    0,
                    attn_weight_outdim // 3,
                ),
                k=QuantInput(
                    attn.c_attn.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.attn.c_attn",
                    attn_weight_outdim // 3,
                    attn_weight_outdim // 3 * 2,
                ),
                v=QuantInput(
                    attn.c_attn.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.attn.c_attn",
                    attn_weight_outdim // 3 * 2,
                    attn_weight_outdim,
                ),
                attn_fc=QuantInput(
                    attn.c_proj.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.attn.c_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.mlp.c_fc",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight.transpose(0, 1),
                    f"{self.quantized_layer_prefix}{index}.mlp.c_proj",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in GPT2LMHeadModel."""
        return (Conv1D,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.attn.c_proj

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.mlp.c_proj

    def get_tf_blocks(self, model: GPT2LMHeadModel) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.transformer.h)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli MPTForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List, Tuple, Type

import torch

from friendli.enums import ModelDataType
from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.quantizer.awq.base import AWQHook
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.utils import scale_reshape


class AWQMPTHook(AWQHook):
    """AWQ Hook for MPTForCausalLM."""

    def add_pre_scaler(self, model: torch.nn.Module) -> torch.nn.Module:
        """Adds scaler to MPTForCausalLM."""
        for tf_block in self.get_tf_blocks(model):
            attn_fc_scaler = self._register_pre_scaler(
                tf_block.attn.out_proj,
            )
            tf_block.attn.add_module("scaler", attn_fc_scaler)
            ff2_scaler = self._register_pre_scaler(tf_block.ffn.down_proj)
            tf_block.ffn.add_module("scaler", ff2_scaler)
        return model

    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[Type[torch.nn.Module], ...]:
        """Returns the type of linear layer (etc. qkv, linear layer) in transformer block."""
        return (type(block.attn), type(block.ffn))

    def iter_inspect_modules(
        self,
        block: torch.nn.Module,
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[Tuple[ModuleName, torch.nn.Linear]],
            torch.nn.Module,
            ModuleName,
        ]
    ]:
        """Returns iterator of layers in modules."""
        # qkv proj
        yield (
            [block.norm_1],
            [("attn.Wqkv", block.attn.Wqkv)],
            block.attn,
            "attn",
        )
        # attn out proj
        yield (
            [block.attn.scaler],
            [("attn.out_proj", block.attn.out_proj)],
            block.attn.out_proj,
            "attn.out_proj",
        )
        # ff1
        yield (
            [block.norm_2],
            [("ffn.up_proj", block.ffn.up_proj)],
            block.ffn,
            "ffn",
        )
        # ff2
        yield (
            [block.ffn.scaler],
            [("ffn.down_proj", block.ffn.down_proj)],
            block.ffn.down_proj,
            "ffn.down_proj",
        )

    def iter_tf_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of MPTForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.attn
            q_outdim = (
                self.converter.decoder_num_attention_heads
                * self.converter.decoder_head_size
            )
            kv_outdim = (
                self.converter.decoder_num_kv_attention_heads
                * self.converter.decoder_head_size
            )
            qkv_outdim = self_attn.Wqkv.weight.size(0)
            assert qkv_outdim == q_outdim + kv_outdim * 2
            fc1 = decoder_layer.ffn.up_proj  # type: ignore
            fc2 = decoder_layer.ffn.down_proj  # type: ignore

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    0,
                    q_outdim,
                ),
                k=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    q_outdim,
                    q_outdim + kv_outdim,
                ),
                v=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    q_outdim + kv_outdim,
                    qkv_outdim,
                ),
                attn_fc=QuantInput(
                    self_attn.out_proj.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.ffn.up_proj",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}{index}.ffn.down_proj",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in MPTForCausalLM."""
        return (torch.nn.Linear,)

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in MPTForCausalLM."""
        return model.transformer.blocks  # type: ignore

    @property
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified layers."""
        convert_info_list = []
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn.scaler.scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}attn/c_proj/awq/pre_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ffn.scaler.scale"],
                        data_type=ModelDataType.FP32,
                        converted_name=f"{converted_prefix}mlp/c_proj/awq/pre_scale:0",
                        reshape_fn=scale_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def avoid_clipping_layer_names(self) -> List[str]:
        """Returns the layer names which should be avoided for AWQ clipping."""
        return ["Wqkv"]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-locals, redefined-builtin, too-many-statements, too-many-arguments

"""Friendli Endpoint CLI."""

from __future__ import annotations

import typer

from friendli.formatter import PanelFormatter, TableFormatter
from friendli.sdk.client import Friendli
from friendli.utils.compat import model_dump
from friendli.utils.decorator import check_api

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)

DEFAULT_MAX_BATCH_SIZE = 256
DEFAULT_MAX_TOKEN_COUNT = 8192

panel_formatter = PanelFormatter(
    name="Endpoint Overview",
    fields=[
        "id",
        "name",
        "hfModelRepo",
        "gpuType",
        "numGpu",
        "status",
        "phase.step",
        "endpointUrl",
        "phase.currReplica",
        "phase.desiredReplica",
        "createdBy.email",
        "createdAt",
        "updatedAt",
    ],
    headers=[
        "ID",
        "Name",
        "Model",
        "GPU Type",
        "GPU Count",
        "Status",
        "Detailed Status",
        "Endpoint URL",
        "Current # Replicas",
        "Desired # Replicas",
        "Created By",
        "Created At",
        "Updated At",
    ],
    substitute_exact_match_only=False,
)

table_formatter = TableFormatter(
    name="Endpoints",
    fields=[
        "id",
        "name",
        "hfModelRepo",
        "gpuType",
        "numGpu",
        "status",
        "createdAt",
    ],
    headers=[
        "ID",
        "Name",
        "Model",
        "GPU Type",
        "GPU Count",
        "Status",
        "Created At",
    ],
    substitute_exact_match_only=False,
)

panel_formatter.add_substitution_rule(
    "INITIALIZING", "[bold yellow]INITIALIZING[/bold yellow]"
)
panel_formatter.add_substitution_rule("RUNNING", "[bold green]RUNNING[/bold green]")
panel_formatter.add_substitution_rule("SLEEPING", "[bold cyan]SLEEPING[/bold cyan]")
panel_formatter.add_substitution_rule(
    "STOPPING", "[bold magenta]STOPPING[/bold magenta]"
)
panel_formatter.add_substitution_rule("FAILED", "[bold red]FAILED[/bold red]")
panel_formatter.add_substitution_rule("TERMINATED", "[bold]TERMINATED[/bold]")

table_formatter.add_substitution_rule(
    "INITIALIZING", "[bold yellow]INITIALIZING[/bold yellow]"
)
table_formatter.add_substitution_rule("RUNNING", "[bold green]RUNNING[/bold green]")
table_formatter.add_substitution_rule("SLEEPING", "[bold cyan]SLEEPING[/bold cyan]")
table_formatter.add_substitution_rule(
    "STOPPING", "[bold magenta]STOPPING[/bold magenta]"
)
table_formatter.add_substitution_rule("FAILED", "[bold red]FAILED[/bold red]")
table_formatter.add_substitution_rule("TERMINATED", "[bold]TERMINATED[/bold]")


@app.command()
@check_api
def create(
    name: str = typer.Option(
        ..., "--name", "-n", help="The name of endpoint to create."
    ),
    model_repo: str = typer.Option(
        ..., "--model", "-m", help="The name of Hugging Face model to deploy."
    ),
    gpu_type: str = typer.Option(
        ..., "--gpu-type", "-gt", help="GPU type to serve the deployed model."
    ),
    gpu_count: int = typer.Option(..., "--gpu-count", "-gc"),
):
    """Creates a new endpoint with deploying model."""
    client = Friendli()

    endpoint = client.endpoint.create(
        name=name,
        model_repo=model_repo,
        gpu_type=gpu_type,
        num_gpus=gpu_count,
    )
    panel_formatter.render(model_dump(endpoint))


@app.command("list")
@check_api
def list_endpoints():
    """List endpoints."""
    client = Friendli()
    endpoints = client.endpoint.list()
    endpoints_ = [model_dump(endpoint) for endpoint in iter(endpoints)]
    table_formatter.render(endpoints_)


@app.command()
@check_api
def get(endpoint_id: str = typer.Argument(..., help="ID of an endpoint to get.")):
    """Get a detailed info of an endpoint."""
    client = Friendli()
    endpoint = client.endpoint.get(endpoint_id)
    panel_formatter.render(model_dump(endpoint))


@app.command()
@check_api
def terminate(
    endpoint_id: str = typer.Argument(..., help="ID of an endpoint to terminate.")
):
    """Terminate a running endpoint."""
    client = Friendli()
    client.endpoint.terminate(endpoint_id=endpoint_id)
    typer.secho(
        f"Endpoint '{endpoint_id}' is terminated successfully.", fg=typer.colors.GREEN
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Defining Friendli Checkpoint Converter maps."""

from __future__ import annotations

from typing import Dict, Tuple, Type, Union

from transformers import (  # type: ignore[import]
    AutoModelForCausalLM,
    BlenderbotForConditionalGeneration,
    BloomForCausalLM,
    CodeGenForCausalLM,
    FalconForCausalLM,
    GPT2LMHeadModel,
    GPTJForCausalLM,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    MistralForCausalLM,
    MixtralForCausalLM,
    MptForCausalLM,
    OPTForCausalLM,
    PreTrainedModel,
    T5ForConditionalGeneration,
)

from friendli.errors import NotSupportedCheckpointError
from friendli.modules.converter.base import OneOfAdapterConverter, OneOfConverter
from friendli.modules.converter.models.blenderbot import BlenderbotConverter
from friendli.modules.converter.models.bloom import BloomForCausalLMConverter
from friendli.modules.converter.models.codegen import CodegenForCausalLMConverter
from friendli.modules.converter.models.falcon import FalconForCausalLMConverter
from friendli.modules.converter.models.gpt2 import GPT2LMHeadModelConverter
from friendli.modules.converter.models.gpt_neox import GPTNeoXForCausalLMConverter
from friendli.modules.converter.models.gptj import (
    GPTJForCausalLMConverter,
    GPTJForCausalLMLoraConverter,
)
from friendli.modules.converter.models.llama import (
    LlamaForCausalLMConverter,
    LlamaForCausalLMLoraConverter,
)
from friendli.modules.converter.models.mistral import MistralForCausalLMConverter
from friendli.modules.converter.models.mixtral import MixtralForCausalLMConverter
from friendli.modules.converter.models.mpt import (
    MPTForCausalLMConverter,
    MptForCausalLMLoraConverter,
)
from friendli.modules.converter.models.opt import OPTForCausalLMConverter
from friendli.modules.converter.models.phi_msft import PhiForCausalLMConverter
from friendli.modules.converter.models.t5 import T5Converter

MODEL_ARCH_CONVERTER_MAP: Dict[
    str, Tuple[Union[PreTrainedModel, PreTrainedModel], Type[OneOfConverter]]
] = {
    "BlenderbotForConditionalGeneration": (
        BlenderbotForConditionalGeneration,
        BlenderbotConverter,
    ),
    "BloomForCausalLM": (BloomForCausalLM, BloomForCausalLMConverter),
    "CodeGenForCausalLM": (CodeGenForCausalLM, CodegenForCausalLMConverter),
    "FalconForCausalLM": (FalconForCausalLM, FalconForCausalLMConverter),
    "GPTNeoXForCausalLM": (GPTNeoXForCausalLM, GPTNeoXForCausalLMConverter),
    "GPT2LMHeadModel": (GPT2LMHeadModel, GPT2LMHeadModelConverter),
    "GPTJForCausalLM": (GPTJForCausalLM, GPTJForCausalLMConverter),
    "LlamaForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "LLaMAForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "MistralForCausalLM": (MistralForCausalLM, MistralForCausalLMConverter),
    "MixtralForCausalLM": (MixtralForCausalLM, MixtralForCausalLMConverter),
    "MPTForCausalLM": (MptForCausalLM, MPTForCausalLMConverter),
    "OPTForCausalLM": (OPTForCausalLM, OPTForCausalLMConverter),
    "T5ForConditionalGeneration": (T5ForConditionalGeneration, T5Converter),
    "PhiForCausalLM": (AutoModelForCausalLM, PhiForCausalLMConverter),
}

MODEL_ARCH_ADAPTER_CONVERTER_MAP: Dict[
    str,
    Type[OneOfAdapterConverter],
] = {
    "GPTJForCausalLM": GPTJForCausalLMLoraConverter,
    "LlamaForCausalLM": LlamaForCausalLMLoraConverter,
    "LLaMAForCausalLM": LlamaForCausalLMLoraConverter,
    "MPTForCausalLM": MptForCausalLMLoraConverter,
}


def get_hf_converter_factory(
    model_arch: str,
) -> Tuple[PreTrainedModel, Type[OneOfConverter]]:
    """Return the converter factory for the given model architecture.

    Args:
        model_arch (str): Model architecture name.

    Returns:
        Tuple[PretrainedModel, Type[OneOfConverter]]: Tuple of
            model class and converter class.

    Raises:
        NotSupportedCheckpointError: Raised when the given model architecture is not supported.

    """
    if model_arch not in MODEL_ARCH_CONVERTER_MAP:
        raise NotSupportedCheckpointError(
            invalid_option=f"Model architecture='{model_arch}'",
            valid_options=list(MODEL_ARCH_CONVERTER_MAP.keys()),
        )

    return MODEL_ARCH_CONVERTER_MAP[model_arch]


def get_adapter_converter_factory(
    model_arch: str,
) -> Type[OneOfAdapterConverter]:
    """Return the converter factory for the given model architecture.

    Args:
        model_arch (str): Model architecture name.

    Returns:
        Type[LoraConverter]: Adapter Converter class.

    Raises:
        NotSupportedCheckpointError: Raised when the given model architecture is not supported.
    """
    try:
        adapter_converter_type = MODEL_ARCH_ADAPTER_CONVERTER_MAP[model_arch]
    except KeyError as exc:
        raise NotSupportedCheckpointError(
            invalid_option=f"adapter for model architecture='{model_arch}'",
            valid_options=list(MODEL_ARCH_ADAPTER_CONVERTER_MAP.keys()),
        ) from exc
    return adapter_converter_type



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI File System Management Utilities."""

from __future__ import annotations

import os
import re
import zipfile
from contextlib import contextmanager
from datetime import datetime
from io import BufferedReader
from pathlib import Path
from typing import Any, Dict, Iterator, List, Union

import pathspec  # type: ignore
import typer
from dateutil.tz import tzlocal

friendli_directory = Path.home() / ".friendli"


def get_friendli_directory() -> Path:
    """Get a hidden path to Friendli config directory."""
    friendli_directory.mkdir(exist_ok=True)
    return friendli_directory


@contextmanager
def zip_dir(
    base_path: Path, target_files: List[Path], zip_path: Path
) -> Iterator[BufferedReader]:
    """Zip direcotry.

    Args:
        base_path (Path): Base directory path. The target files will be located at the
            path related to the base path when the zip file is unzipped.
        target_files (List[Path]): All target files to zip.
        zip_path (Path): Path to save zip file.

    Yields:
        BufferedReader: Opened zip file object.

    """
    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_STORED) as zip_file:
        for file in target_files:
            zip_file.write(file, file.relative_to(base_path.parent))
    typer.secho("Uploading workspace directory...", fg=typer.colors.MAGENTA)
    try:
        yield zip_path.open("rb")
    finally:
        zip_path.unlink()


def get_workspace_files(dir_path: Path) -> List[Path]:
    """Get workspace file paths.

    Args:
        dir_path (Path): Path to the workspace directory.

    Returns:
        List[Path]: A list of paths to the file inside the workspace.

    """
    ignore_file = dir_path / ".pfignore"
    all_files = set(x for x in dir_path.rglob("*") if x.is_file() and x != ignore_file)

    if not ignore_file.exists():
        return list(all_files)

    with open(ignore_file, "r", encoding="utf-8") as f:
        ignore_patterns = f.read()
    spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern, ignore_patterns.splitlines()
    )
    matched_files = set(dir_path / x for x in spec.match_tree_files(dir_path))  # type: ignore
    return list(all_files.difference(matched_files))


def storage_path_to_local_path(storage_path: str, source_path: Path) -> Path:
    """Translate a cloud storage path to the local file system path."""
    return Path(strip_storage_path_prefix(str(source_path / storage_path)))


def get_file_info(storage_path: str, source_path: Path) -> Dict[str, Any]:
    """Get file metadata."""
    local_path = storage_path_to_local_path(storage_path, source_path)
    stat = local_path.stat()
    return {
        "name": os.path.basename(storage_path),
        "path": storage_path,
        "mtime": datetime.fromtimestamp(stat.st_mtime, tz=tzlocal()).isoformat(),
        "size": stat.st_size,
    }


def get_file_size(file_path: Union[str, Path]) -> int:
    """Calculate a file size in bytes.

    Args:
        file_path (Union[str, Path]): Path to the target file.

    Returns:
        int: The size of a file.

    """
    if isinstance(file_path, str):
        return os.stat(file_path).st_size
    return file_path.stat().st_size


def strip_storage_path_prefix(path: str) -> str:
    """Strip a prefix about auxiliary checkpoint info from the path.

    Args:
        path (str): Actual checkpoint storage path. The path may starts with the prefix
            that contain the iteration number and distributed training configuration.

    Returns:
        str: Path without the prefix.

    """
    return re.sub(
        pattern=r"iter_\d{7}/mp\d{3}-\d{3}pp\d{3}-\d{3}/",
        repl="",
        string=path,
    )


def attach_storage_path_prefix(
    path: Union[str, Path],
    iteration: int,
    mp_rank: int,
    mp_degree: int,
    pp_rank: int,
    pp_degree: int,
) -> str:
    """Attach a filename prefix to mark up iteration and dist info.

    Args:
        path (Union[str, Path]): The relative path to file.
        iteration (int): Checkpoint iteration number.
        mp_rank (int): Model parallelism(a.k.a. tensor parallelism) rank.
        mp_degree (int): Model parallelism(a.k.a. tensor parallelism) degree.
        pp_rank (int): Pipelined model parallelism rank.
        pp_degree (int): Pipelined model parallelism degree.

    Returns:
        str: Storage path with the prefix

    """
    # pylint: disable=line-too-long
    return f"iter_{iteration:07d}/mp{mp_rank:03d}-{mp_degree:03d}pp{pp_rank:03d}-{pp_degree:03d}/{path}"



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli Client."""

from __future__ import annotations

from typing import Optional

import friendli
from friendli.client.graphql.endpoint import EndpointGqlClient
from friendli.client.graphql.model import ModelGqlClient
from friendli.sdk.api.chat.chat import AsyncChat, Chat
from friendli.sdk.api.completions import AsyncCompletions, Completions
from friendli.sdk.api.images.images import AsyncImages, Images
from friendli.sdk.resource.endpoint import EndpointApi
from friendli.sdk.resource.model import ModelApi

INFERENCE_ENDPOINT_URL = "https://inference.friendli.ai"


class FriendliClientBase:
    """Base class of Friendli client."""

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        endpoint_id: Optional[str] = None,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
    ):
        """Initializes FriendliClientBase."""
        if token is not None:
            friendli.token = token
        if team_id is not None:
            friendli.team_id = team_id
        if project_id is not None:
            friendli.project_id = project_id
        self._endpoint_id = endpoint_id
        self._base_url = base_url
        self._use_protobuf = use_protobuf


class Friendli(FriendliClientBase):
    """Friendli API client."""

    completions: Completions
    chat: Chat
    images: Images
    endpoint: EndpointApi
    model: ModelApi

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        endpoint_id: Optional[str] = None,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
    ):
        """Initializes Friendli."""
        super().__init__(
            token=token,
            team_id=team_id,
            project_id=project_id,
            endpoint_id=endpoint_id,
            base_url=base_url,
            use_protobuf=use_protobuf,
        )

        base_url = base_url or INFERENCE_ENDPOINT_URL
        self.completions = Completions(
            base_url=base_url, endpoint_id=self._endpoint_id, use_protobuf=use_protobuf
        )
        self.chat = Chat(
            base_url=base_url, endpoint_id=self._endpoint_id, use_protobuf=use_protobuf
        )
        self.images = Images(base_url=base_url, endpoint_id=self._endpoint_id)

        endpoint_client = EndpointGqlClient()
        model_client = ModelGqlClient()
        self.endpoint = EndpointApi(client=endpoint_client)
        self.model = ModelApi(client=model_client)


class AsyncFriendli(FriendliClientBase):
    """Async Friendli API client."""

    completions: AsyncCompletions
    chat: AsyncChat
    images: AsyncImages

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        endpoint_id: Optional[str] = None,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
    ):
        """Initializes AsyncFriendli."""
        super().__init__(
            token=token,
            team_id=team_id,
            project_id=project_id,
            endpoint_id=endpoint_id,
            base_url=base_url,
            use_protobuf=use_protobuf,
        )

        base_url = base_url or INFERENCE_ENDPOINT_URL
        self.completions = AsyncCompletions(
            base_url=base_url, endpoint_id=self._endpoint_id, use_protobuf=use_protobuf
        )
        self.chat = AsyncChat(
            base_url=base_url, endpoint_id=self._endpoint_id, use_protobuf=use_protobuf
        )
        self.images = AsyncImages(base_url=base_url, endpoint_id=self._endpoint_id)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli GPTNeoXForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Tuple, Type, cast

import torch
from transformers.models.gpt_neox import (  # type: ignore[import]
    GPTNeoXConfig,
    GPTNeoXForCausalLM,
)

from friendli.modules.converter.base import OneOfConverter
from friendli.modules.converter.utils import convert_to_gpt_j_params
from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantGPTNeoXHook(SmoothQuantHook):
    """SmoothQuant Hook for GPTNeoXForCausalLM."""

    def __init__(self, quant_config: Dict[str, Any], converter: OneOfConverter):
        """Initialize SmoothQuantGPTNeoXHook."""
        super().__init__(quant_config, converter)
        config = cast(GPTNeoXConfig, converter.config)
        self.num_attention_heads = config.num_attention_heads
        self.num_kv_attention_heads = config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.head_size = self.hidden_size // self.num_attention_heads
        self.rotary_dim = int(self.head_size * config.rotary_pct)

    def iter_smooth_norm_weights(
        self,
        model: GPTNeoXForCausalLM,
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in GPTNeoXForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.gpt_neox.layers):  # type: ignore[union-attr]
            # [LayerNorm 1] - [ QKV projection ] gets smoothed
            yield (
                [
                    decoder_layer.input_layernorm.weight.data,
                    decoder_layer.input_layernorm.bias.data,
                ],
                [
                    decoder_layer.attention.query_key_value.weight.data,
                ],
                f"{self.quantized_layer_prefix}{index}.attention.query_key_value",  # the input tensors fed into Q, K, V matrices are identical.
            )
            # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
            yield (
                [
                    decoder_layer.post_attention_layernorm.weight.data,
                    decoder_layer.post_attention_layernorm.bias.data,
                ],
                [decoder_layer.mlp.dense_h_to_4h.weight.data],  # [OutDim, InDim]
                f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.attention.dense.weight.data],
                    f"{self.quantized_layer_prefix}{index}.attention.dense",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.mlp.dense_4h_to_h.weight.data],
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                )

    def reshape_qkv_weight(
        self, attn_layer: torch.nn.Module
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Reshape GPTNeoXForCausalLM's qkv weight for int8 quantization."""
        qkv_weight = cast(torch.nn.Linear, attn_layer).weight
        qkv_weight = qkv_weight.reshape(
            self.num_attention_heads,
            3,
            self.head_size,
            self.hidden_size,
        )

        q_weight = qkv_weight[:, 0].reshape(
            self.num_attention_heads,
            self.head_size,
            self.hidden_size,
        )
        k_weight = qkv_weight[:, 1].reshape(
            self.num_attention_heads,
            self.head_size,
            self.hidden_size,
        )
        v_weight = qkv_weight[:, 2].reshape(
            self.num_attention_heads * self.head_size,
            self.hidden_size,
        )

        q_weight = convert_to_gpt_j_params(param=q_weight, rotary_dim=self.rotary_dim)
        k_weight = convert_to_gpt_j_params(param=k_weight, rotary_dim=self.rotary_dim)
        q_weight = q_weight.reshape(
            self.num_attention_heads * self.head_size,
            self.hidden_size,
        )
        k_weight = k_weight.reshape(
            self.num_attention_heads * self.head_size,
            self.hidden_size,
        )
        return q_weight, k_weight, v_weight

    def sort_qkv_output_stats(self, max_output_stat: torch.Tensor) -> torch.Tensor:
        """Sort max output stats of qkv_layer in GPTNeoXForCausalLM."""
        max_output_stat = max_output_stat.reshape(
            self.num_attention_heads,
            3,
            self.head_size,
        )
        q_output_stat = max_output_stat[:, 0].reshape(
            self.num_attention_heads,
            self.head_size,
        )
        k_output_stat = max_output_stat[:, 1].reshape(
            self.num_attention_heads,
            self.head_size,
        )
        v_output_stat = max_output_stat[:, 2].reshape(
            self.num_attention_heads * self.head_size,
        )
        q_output_stat = convert_to_gpt_j_params(q_output_stat, self.rotary_dim)
        k_output_stat = convert_to_gpt_j_params(k_output_stat, self.rotary_dim)
        q_output_stat = q_output_stat.reshape(
            self.num_attention_heads * self.head_size,
        )
        k_output_stat = k_output_stat.reshape(
            self.num_attention_heads * self.head_size,
        )
        return torch.cat((q_output_stat, k_output_stat, v_output_stat), dim=0)

    def iter_tf_quant_inputs(
        self, model: GPTNeoXForCausalLM
    ) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of GPTNeoXForCausalLM."""
        for index, decoder_layer in enumerate(model.gpt_neox.layers):
            attention = decoder_layer.attention
            attention_weight_outdim = attention.query_key_value.weight.size(0)  # OutDim
            q_weight, k_weight, v_weight = self.reshape_qkv_weight(
                attention.query_key_value
            )
            qkv_weight = torch.cat((q_weight, k_weight, v_weight), dim=0)
            fc1 = decoder_layer.mlp.dense_h_to_4h
            fc2 = decoder_layer.mlp.dense_4h_to_h

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.attention.query_key_value",
                    0,
                    attention_weight_outdim // 3,
                    self.sort_qkv_output_stats,
                ),
                k=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.attention.query_key_value",
                    attention_weight_outdim // 3,
                    attention_weight_outdim // 3 * 2,
                    self.sort_qkv_output_stats,
                ),
                v=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.attention.query_key_value",
                    attention_weight_outdim // 3 * 2,
                    attention_weight_outdim,
                    self.sort_qkv_output_stats,
                ),
                attn_fc=QuantInput(
                    attention.dense.weight,
                    f"{self.quantized_layer_prefix}{index}.attention.dense",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in GPTNeoXForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.attention.dense

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.mlp.dense_4h_to_h

    def get_tf_blocks(self, model: GPTNeoXForCausalLM) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.gpt_neox.layers)
